@@ -173,23 +173,19 @@ static struct {
 #endif
 
 #ifdef COG_USE_WAYLAND_CURSOR
-    struct wl_shm          *wl_shm;
     struct wl_cursor_theme *cursor_theme;
     struct wl_cursor       *cursor_left_ptr;
     struct wl_surface      *cursor_left_ptr_surface;
 #endif /* COG_USE_WAYLAND_CURSOR */
 
     struct output_metrics metrics[16];
+    struct output_metrics *current_output;
 
     struct zwp_text_input_manager_v3 *text_input_manager;
     struct zwp_text_input_manager_v1 *text_input_manager_v1;
 
     struct wp_presentation *presentation;
 
-    struct {
-        int32_t scale;
-    } current_output;
-
     struct {
         struct wl_pointer *obj;
         struct wl_surface *surface;
@@ -231,9 +227,7 @@ static struct {
     GSource *event_src;
 
     struct wl_list shm_buffer_list;
-} wl_data = {
-    .current_output.scale = 1,
-};
+} wl_data = {};
 
 static struct {
     struct egl_display *display;
@@ -442,12 +436,16 @@ configure_surface_geometry(int32_t width, int32_t height)
 static void
 resize_window(void)
 {
-    int32_t pixel_width = win_data.width * wl_data.current_output.scale;
-    int32_t pixel_height = win_data.height * wl_data.current_output.scale;
+    int32_t pixel_width = win_data.width * wl_data.current_output->scale;
+    int32_t pixel_height = win_data.height * wl_data.current_output->scale;
 
     wpe_view_backend_dispatch_set_size(wpe_view_data.backend, win_data.width, win_data.height);
-    g_debug("Resized EGL buffer to: (%u, %u) @%ix\n", pixel_width, pixel_height, wl_data.current_output.scale);
+    wpe_view_backend_dispatch_set_device_scale_factor(wpe_view_data.backend, wl_data.current_output->scale);
+
+    g_debug("Resized EGL buffer to: (%" PRIi32 ", %" PRIi32 ") @%" PRIi32 "x", pixel_width, pixel_height,
+            wl_data.current_output->scale);
 }
+
 static void
 shell_surface_ping(void *data, struct wl_shell_surface *shell_surface, uint32_t serial)
 {
@@ -619,6 +617,23 @@ output_handle_mode(void *data, struct wl_output *output, uint32_t flags, int32_t
 static void
 output_handle_done(void *data, struct wl_output *output)
 {
+    struct output_metrics *metrics = find_output(output);
+
+    if (!metrics->refresh) {
+        g_warning("No refresh rate reported for output %p, using 60Hz", output);
+        metrics->refresh = 60 * 1000;
+    }
+
+    if (!metrics->scale) {
+        g_warning("No scale factor reported for output %p, using 1x", output);
+        metrics->scale = 1;
+    }
+
+    if (!wl_data.current_output) {
+        g_debug("%s: Using %p as initial output", G_STRFUNC, output);
+        wl_data.current_output = metrics;
+    }
+
     if (win_data.should_resize_to_largest_output) {
         resize_to_largest_output();
     }
@@ -634,7 +649,7 @@ output_handle_scale(void *data, struct wl_output *output, int32_t factor)
     }
 
     metrics->scale = factor;
-    g_info("Got scale factor %i for output %p\n", factor, output);
+    g_info("Got scale factor %i for output %p", factor, output);
 }
 #endif /* WL_OUTPUT_SCALE_SINCE_VERSION */
 
@@ -733,33 +748,23 @@ static const struct wl_output_listener output_listener = {
 static void
 surface_handle_enter(void *data, struct wl_surface *surface, struct wl_output *output)
 {
-#ifdef WL_SURFACE_SET_BUFFER_SCALE_SINCE_VERSION
-    int32_t scale_factor = -1;
-    int32_t refresh = 0;
+    if (wl_data.current_output->output != output) {
+        g_debug("%s: Surface %p output changed %p -> %p", G_STRFUNC, surface, wl_data.current_output->output, output);
+        wl_data.current_output = find_output(output);
+        g_assert(wl_data.current_output);
+    }
 
-    for (int i = 0; i < G_N_ELEMENTS(wl_data.metrics); i++) {
-        if (wl_data.metrics[i].output == output) {
-            scale_factor = wl_data.metrics[i].scale;
-            refresh = wl_data.metrics[i].refresh;
-        }
-    }
-    if (scale_factor == -1) {
-        g_warning("No scale factor available for output %p", output);
-        return;
-    }
-    if (refresh < 0) {
-        g_warning("Refresh rate less than zero for output %p", output);
-        return;
-    }
+#ifdef WL_SURFACE_SET_BUFFER_SCALE_SINCE_VERSION
     if (wl_surface_get_version(surface) >= WL_SURFACE_SET_BUFFER_SCALE_SINCE_VERSION) {
-        wl_surface_set_buffer_scale(surface, scale_factor);
-        wpe_view_backend_dispatch_set_device_scale_factor(wpe_view_data.backend, scale_factor);
-        wl_data.current_output.scale = scale_factor;
+        wl_surface_set_buffer_scale(surface, wl_data.current_output->scale);
+        wpe_view_backend_dispatch_set_device_scale_factor(wpe_view_data.backend, wl_data.current_output->scale);
+    } else {
+        g_debug("%s: Surface %p uses old protocol version, cannot set scale factor", G_STRFUNC, surface);
     }
 #endif /* WL_SURFACE_SET_BUFFER_SCALE_SINCE_VERSION */
 
 #if HAVE_REFRESH_RATE_HANDLING
-    wpe_view_backend_set_target_refresh_rate(wpe_view_data.backend, refresh);
+    wpe_view_backend_set_target_refresh_rate(wpe_view_data.backend, wl_data.current_output->refresh);
 #endif /* HAVE_REFRESH_RATE_HANDLING */
 }
 
@@ -886,14 +891,12 @@ pointer_on_motion(void *data, struct wl_pointer *pointer, uint32_t time, wl_fixe
     wl_data.pointer.x = wl_fixed_to_int(fixed_x);
     wl_data.pointer.y = wl_fixed_to_int(fixed_y);
 
-    struct wpe_input_pointer_event event = {
-        wpe_input_pointer_event_type_motion,
-        time,
-        wl_data.pointer.x * wl_data.current_output.scale,
-        wl_data.pointer.y * wl_data.current_output.scale,
-        wl_data.pointer.button,
-        wl_data.pointer.state
-    };
+    struct wpe_input_pointer_event event = {wpe_input_pointer_event_type_motion,
+                                            time,
+                                            wl_data.pointer.x * wl_data.current_output->scale,
+                                            wl_data.pointer.y * wl_data.current_output->scale,
+                                            wl_data.pointer.button,
+                                            wl_data.pointer.state};
 
     wpe_view_backend_dispatch_pointer_event(wpe_view_data.backend, &event);
 }
@@ -921,8 +924,8 @@ pointer_on_button(void *data,
     struct wpe_input_pointer_event event = {
         wpe_input_pointer_event_type_button,
         time,
-        wl_data.pointer.x * wl_data.current_output.scale,
-        wl_data.pointer.y * wl_data.current_output.scale,
+        wl_data.pointer.x * wl_data.current_output->scale,
+        wl_data.pointer.y * wl_data.current_output->scale,
         wl_data.pointer.button,
         wl_data.pointer.state,
     };
@@ -952,11 +955,11 @@ dispatch_axis_event()
     struct wpe_input_axis_2d_event event = { 0, };
     event.base.type = wpe_input_axis_event_type_mask_2d | wpe_input_axis_event_type_motion_smooth;
     event.base.time = wl_data.axis.time;
-    event.base.x = wl_data.pointer.x * wl_data.current_output.scale;
-    event.base.y = wl_data.pointer.y * wl_data.current_output.scale;
+    event.base.x = wl_data.pointer.x * wl_data.current_output->scale;
+    event.base.y = wl_data.pointer.y * wl_data.current_output->scale;
 
-    event.x_axis = wl_fixed_to_double(wl_data.axis.x_delta) * wl_data.current_output.scale;
-    event.y_axis = -wl_fixed_to_double(wl_data.axis.y_delta) * wl_data.current_output.scale;
+    event.x_axis = wl_fixed_to_double(wl_data.axis.x_delta) * wl_data.current_output->scale;
+    event.y_axis = -wl_fixed_to_double(wl_data.axis.y_delta) * wl_data.current_output->scale;
 
     wpe_view_backend_dispatch_axis_event(wpe_view_data.backend, &event.base);
 
@@ -975,23 +978,28 @@ pointer_uses_frame_event(struct wl_pointer *pointer)
 #endif
 }
 
+/*
+ * Wayland reports axis events as being 15 units per scroll wheel step. Scale
+ * values in order to match the 120 value used in the X11 plug-in and by the
+ * libinput_event_pointer_get_scroll_value_v120() function.
+ */
+enum {
+    SCROLL_WHEEL_STEP_SCALING_FACTOR = 8,
+};
+
 static void
-pointer_on_axis (void* data,
-                 struct wl_pointer *pointer,
-                 uint32_t time,
-                 uint32_t axis,
-                 wl_fixed_t value)
+pointer_on_axis(void *data, struct wl_pointer *pointer, uint32_t time, uint32_t axis, wl_fixed_t value)
 {
     if (axis == WL_POINTER_AXIS_VERTICAL_SCROLL) {
         wl_data.axis.has_delta = true;
         wl_data.axis.time = time;
-        wl_data.axis.y_delta += value;
+        wl_data.axis.y_delta += value * SCROLL_WHEEL_STEP_SCALING_FACTOR;
     }
 
     if (axis == WL_POINTER_AXIS_HORIZONTAL_SCROLL) {
         wl_data.axis.has_delta = true;
         wl_data.axis.time = time;
-        wl_data.axis.x_delta += value;
+        wl_data.axis.x_delta += value * SCROLL_WHEEL_STEP_SCALING_FACTOR;
     }
 
     if (!pointer_uses_frame_event(pointer))
@@ -1177,10 +1185,9 @@ handle_key_event(CogWlPlatform *self, uint32_t key, uint32_t state, uint32_t tim
             && xkb_compose_state_get_status (xkb_data.compose_state)
             == XKB_COMPOSE_COMPOSED) {
         keysym = xkb_compose_state_get_one_sym (xkb_data.compose_state);
-        unicode = xkb_keysym_to_utf32 (keysym);
     }
 
-    struct wpe_input_keyboard_event event = {time, keysym, unicode, state == true, xkb_data.modifiers};
+    struct wpe_input_keyboard_event event = {time, keysym, key, state == true, xkb_data.modifiers};
 
     wpe_view_backend_dispatch_keyboard_event (wpe_view_data.backend, &event);
 }
@@ -1209,8 +1216,15 @@ keyboard_on_key (void *data,
 {
     CogWlPlatform *self = data;
 
-    /* @FIXME: investigate why is this necessary */
-    // IDK.
+    // wl_keyboard protocol sends key events as a physical key signals on
+    // the keyboard (limited to 256 - 8 bits).  The XKB protocol doesn't share
+    // this limitation and uses extended keycodes and it restricts these
+    // names to at most 4 (ASCII) characters:
+    //
+    //   xkb_keycode_t keycode_A = KEY_A + 8;
+    //
+    // Ref: xkb_keycode_t section in
+    // https://xkbcommon.org/doc/current/xkbcommon_8h.html
     key += 8;
 
     wl_data.event_serial = serial;
@@ -1330,8 +1344,8 @@ touch_on_down (void *data,
         wpe_input_touch_event_type_down,
         time,
         id,
-        wl_fixed_to_int (x) * wl_data.current_output.scale,
-        wl_fixed_to_int (y) * wl_data.current_output.scale,
+        wl_fixed_to_int(x) * wl_data.current_output->scale,
+        wl_fixed_to_int(y) * wl_data.current_output->scale,
     };
 
     memcpy (&wl_data.touch.points[id],
@@ -1430,8 +1444,8 @@ touch_on_motion (void *data,
         wpe_input_touch_event_type_motion,
         time,
         id,
-        wl_fixed_to_int (x) * wl_data.current_output.scale,
-        wl_fixed_to_int (y) * wl_data.current_output.scale,
+        wl_fixed_to_int(x) * wl_data.current_output->scale,
+        wl_fixed_to_int(y) * wl_data.current_output->scale,
     };
 
     memcpy (&wl_data.touch.points[id],
@@ -1654,6 +1668,20 @@ static const struct wl_buffer_listener dmabuf_buffer_listener = {
 static void
 on_export_wl_egl_image(void *data, struct wpe_fdo_egl_exported_image *image)
 {
+    const uint32_t surface_pixel_width = wl_data.current_output->scale * win_data.width;
+    const uint32_t surface_pixel_height = wl_data.current_output->scale * win_data.height;
+
+    if (surface_pixel_width != wpe_fdo_egl_exported_image_get_width(image) ||
+        surface_pixel_height != wpe_fdo_egl_exported_image_get_height(image)) {
+        g_debug("Exported FDO EGL image size %" PRIu32 "x%" PRIu32 ", does not match surface size %" PRIu32 "x%" PRIu32
+                ", skipping.",
+                wpe_fdo_egl_exported_image_get_width(image), wpe_fdo_egl_exported_image_get_height(image),
+                surface_pixel_width, surface_pixel_width);
+        wpe_view_backend_exportable_fdo_dispatch_frame_complete(wpe_host_data.exportable);
+        wpe_view_backend_exportable_fdo_egl_dispatch_release_exported_image(wpe_host_data.exportable, image);
+        return;
+    }
+
     wpe_view_data.image = image;
 
     if (wpe_view_data.should_update_opaque_region) {
@@ -1682,10 +1710,7 @@ on_export_wl_egl_image(void *data, struct wpe_fdo_egl_exported_image *image)
     wl_buffer_add_listener(wpe_view_data.buffer, &buffer_listener, image);
 
     wl_surface_attach (win_data.wl_surface, wpe_view_data.buffer, 0, 0);
-    wl_surface_damage (win_data.wl_surface,
-                       0, 0,
-                       win_data.width * wl_data.current_output.scale,
-                       win_data.height * wl_data.current_output.scale);
+    wl_surface_damage(win_data.wl_surface, 0, 0, surface_pixel_width, surface_pixel_height);
 
     request_frame ();
 
@@ -1798,6 +1823,21 @@ on_export_shm_buffer (void* data, struct wpe_fdo_shm_exported_buffer* exported_b
     struct wl_resource *exported_resource = wpe_fdo_shm_exported_buffer_get_resource (exported_buffer);
     struct wl_shm_buffer *exported_shm_buffer = wpe_fdo_shm_exported_buffer_get_shm_buffer (exported_buffer);
 
+    const uint32_t surface_pixel_width = wl_data.current_output->scale * win_data.width;
+    const uint32_t surface_pixel_height = wl_data.current_output->scale * win_data.height;
+
+    if (surface_pixel_width != wl_shm_buffer_get_width(exported_shm_buffer) ||
+        surface_pixel_height != wl_shm_buffer_get_height(exported_shm_buffer)) {
+        g_debug("Exported SHM buffer size %" PRIu32 "x%" PRIu32 ", does not match surface size %" PRIu32 "x%" PRIu32
+                ", skipping.",
+                wl_shm_buffer_get_width(exported_shm_buffer), wl_shm_buffer_get_width(exported_shm_buffer),
+                surface_pixel_width, surface_pixel_width);
+        wpe_view_backend_exportable_fdo_dispatch_frame_complete(wpe_host_data.exportable);
+        wpe_view_backend_exportable_fdo_egl_dispatch_release_shm_exported_buffer(wpe_host_data.exportable,
+                                                                                 exported_buffer);
+        return;
+    }
+
     struct shm_buffer *buffer = shm_buffer_for_resource (exported_resource);
     if (!buffer) {
         int32_t width;
@@ -1979,14 +2019,11 @@ init_wayland (GError **error)
     wl_display_roundtrip (wl_data.display);
 
 #if COG_USE_WAYLAND_CURSOR
-    if (wl_data.wl_shm) {
-        if (!(wl_data.cursor_theme = wl_cursor_theme_load (NULL,
-                                                           32,
-                                                           wl_data.wl_shm))) {
-            g_warning ("%s: Could not load cursor theme.", G_STRFUNC);
-        } else if (!(wl_data.cursor_left_ptr =
-                     wl_cursor_theme_get_cursor (wl_data.cursor_theme, "left_ptr"))) {
-            g_warning ("%s: Could not load left_ptr cursor.", G_STRFUNC);
+    if (wl_data.shm) {
+        if (!(wl_data.cursor_theme = wl_cursor_theme_load(NULL, 32, wl_data.shm))) {
+            g_warning("%s: Could not load cursor theme.", G_STRFUNC);
+        } else if (!(wl_data.cursor_left_ptr = wl_cursor_theme_get_cursor(wl_data.cursor_theme, "left_ptr"))) {
+            g_warning("%s: Could not load left_ptr cursor.", G_STRFUNC);
         }
     }
 #endif /* COG_USE_WAYLAND_CURSOR */
@@ -2027,7 +2064,6 @@ clear_wayland (void)
 #ifdef COG_USE_WAYLAND_CURSOR
     g_clear_pointer (&wl_data.cursor_left_ptr_surface, wl_surface_destroy);
     g_clear_pointer (&wl_data.cursor_theme, wl_cursor_theme_destroy);
-    g_clear_pointer (&wl_data.wl_shm, wl_shm_destroy);
 #endif /* COG_USE_WAYLAND_CURSOR */
 
     wl_registry_destroy (wl_data.registry);
@@ -2274,17 +2310,15 @@ create_popup (WebKitOptionMenu *option_menu)
     popup_data.width = win_data.width;
     popup_data.height = cog_popup_menu_get_height_for_option_menu (option_menu);
 
-    popup_data.popup_menu = cog_popup_menu_create (option_menu, wl_data.shm,
-                                                   popup_data.width,
-                                                   popup_data.height,
-                                                   wl_data.current_output.scale);
+    popup_data.popup_menu = cog_popup_menu_create(option_menu, wl_data.shm, popup_data.width, popup_data.height,
+                                                  wl_data.current_output->scale);
 
     popup_data.wl_surface = wl_compositor_create_surface (wl_data.compositor);
     g_assert (popup_data.wl_surface);
 
 #ifdef WL_SURFACE_SET_BUFFER_SCALE_SINCE_VERSION
     if (wl_surface_get_version(popup_data.wl_surface) >= WL_SURFACE_SET_BUFFER_SCALE_SINCE_VERSION)
-        wl_surface_set_buffer_scale(popup_data.wl_surface, wl_data.current_output.scale);
+        wl_surface_set_buffer_scale(popup_data.wl_surface, wl_data.current_output->scale);
 #endif /* WL_SURFACE_SET_BUFFER_SCALE_SINCE_VERSION */
 
     if (wl_data.xdg_shell != NULL) {
